-- Washing machines, wavies, wigglers, etc

module Wavies where

import Monad
import Pan
import Simple (cropUI,rippleUI,swirlUI,tileUI')

wiggleRotateP :: FloatE -> FloatE -> TimeE -> TransformE
wiggleRotateP cycles maxAng t p = rotateP ang p
 where
   ang = maxAng * sin (t + (2 * pi * cycles * (distO p / 100 - 0.5)))

wiggleRotate :: FloatE -> FloatE -> TimeE -> Filter c
wiggleRotate cycles maxAng t im =
  im . wiggleRotateP cycles maxAng t

{-
washer :: FloatE -> FloatE -> FloatE -> FloatE -> FilterBC
washer cycles maxRot stretch motSpeed im t =
  crop (disk 50) $
  uscale stretch $
  wiggleRotate cycles (pi * maxRot) t $
  translate (motSpeed/stretch * t, 0) $
  im

washerUI :: UI FilterBC
washerUI = liftM4 washer
              (slider "cycles" (0,5) 1)
              (slider "max rotation" (0,2) (1/4))
              (slider "scale" (0,2) 0.5)
              (slider "motion" (0,150) 0)

-}

washer :: FloatE -> FloatE -> FloatE -> FilterBC
washer cycles maxRot stretch im t =
  wiggleRotate cycles (pi * maxRot) t $
  uscale stretch $
  im

washerUI :: UI FilterBC
washerUI = liftM3 washer
              (slider "cycles" (0,5) 1)
              (slider "max rotation" (0,2) (1/4))
              (slider "scale" (0,2) 0.5)

-- This one moves the underlying image
washerMovingUI :: UI FilterBC
washerMovingUI = liftM2 h washerUI (slider "motion" (-50,150) 0)
 where
   h washerF mot im t = washerF (translate (mot*t,0) im) t

{-
translateXUI :: UI FilterBC
translateXUI = do mot <- slider "motion" (-50,150) 0
             return $ \ im t -> translate (mot*t, 0) im
-}

-- washing a circle-cropped moving bilerp checker
washerBilerpUI :: UI ImageBC
washerBilerpUI = liftM h washerMovingUI
 where
   h washerF = \ t ->
     crop (disk 50) (washerF tiledBigBilerp t)

washerTiledUI :: UI ImageBC
washerTiledUI = washerMovingUI `ap` tileUI'



wv :: BoolE -> FloatE -> FloatE -> FloatE -> (FloatE -> FloatE) -> ImageC
wv cropped squeeze angle waviness f =
  uscale (50/squeeze) $
  ifE cropped (crop (uscale squeeze udisk)) id $
  rotate angle $
  cwave f waviness cb0

wavesUI = do 
             cropped  <- tick     "cropped" True
             waviness <- slider   "waviness" (50,300) 140
             angle    <- slider   "angle" (-pi/2,pi/2) (-pi/4)
             squeeze  <- slider   "squeeze" (50,300) 210
             f        <- radioBox "Function" [("sin",sin), ("tan",tan)] 0
             return $
               wv cropped squeeze angle waviness f

-- ToDo: use maybeUI above.  Also, factor into component UIs with simpler
-- effects.


-- Polar checker

polarChecker :: IntE -> Region
polarChecker n = checker . sc . toPolar
  where sc (r,a) = (r/10, a * n'/pi)
        n'       = i2f n

polarCheckerUI :: UI Region
polarCheckerUI = liftM polarChecker (islider "alternations" (0,10) 3)

polarCheckerCropSwirlUI :: UI ImageC
polarCheckerCropSwirlUI =
  cropUI `ap` (swirlUI `ap` liftM regionToImageC polarCheckerUI)
--   do n      <- islider "alternations" (1,8) 3
--                     cropF  <- cropUI
--                     swirlF <- swirlUI
--                     return $ cropF (regionToImageC (swirlF (polarChecker n)))


wigglers :: IntE -> FloatE -> FloatE -> FloatE -> FloatE
         -> IntE -> FloatE -> FractionE
         -> TimeE -> ImageC
wigglers ripples depth turning sw ir arms stretch ybShift t =
  rippleRadius ripples depth $
  rotate (turning * t) $
  swirl sw $
  crop (uscale 100 (annulus ir)) $
  crop (radReg arms) $
  uscale stretch $
  (. polarXf (\ (r,a) -> (r + 30*ybShift,a))) $
  ybRings

wigglersUI = do ripples <- islider "ripples" (0,20) 3
                depth   <- slider  "depth" (0,1.5) 0.4
                turning <- slider  "turning" (0,3) 0.3
                sw100   <- slider  "100/swirl" (0,1) 0.5
                ir      <- slider  "inner radius" (0,1) 0.2
                arms    <- islider "arms" (0,20) 6
                stretch <- slider  "scale" (0.1,15) 1.75
                ybShift <- slider "yellow-blue shift" (0,1) 0
                return $ uscale 0.5 .
                  wigglers ripples depth turning (100/sw100) ir
                           arms stretch ybShift


swirlFishUI = swirlUI `ap` (
              rippleUI `ap`
              liftM2 yb
                (slider "yellow-blue shift" (0,1) 0)
                (slider  "stretch" (0.1,15) 1.4) )
  where
    yb ybShift stretch =
      crop (disk 100) $
      uscale stretch $
      --(. polarXf (\ (r,a) -> (r + 30*ybShift,a))) $
      (. radialXf ( + (30*ybShift))) $
      ybRings

------ "Whirling", an alternative to swirling


whirl :: IntE -> Image FloatE
whirl arms = progress . toPolar
 where
   -- toPolar gives angle in [-pi,pi], so add pi.
   progress (r,a) = r + i2f arms * (a+pi)/(2*pi)

-- To do: re-do whirl using wedges and swirl.

whirlR :: IntE -> FractionE -> Region
whirlR steps coverage = uscale 10 ((<* coverage) . fracPart . whirl steps)

whirlUI :: UI Region
whirlUI = liftM2 whirlR
            (islider "arms"      (0,20) 5)
            (slider  "thickness" (0, 1) 0.5)

type Binop c = c -> c -> c

counterTurn :: Binop (Image c) -> FloatE -> Filter c
counterTurn combine ang im = im' `combine` vflip im'
 where
   im' = rotate ang im

counterSpinUI :: Binop (Image c) -> UI (Image c -> ImageB c)
counterSpinUI combine = do
  angF  <- liftM (\ r t -> -r*t) (slider "rate" (-5,5) 2)
  return $ \ im t -> counterTurn combine (angF t) im

counterSpinWhirlUI :: UI (ImageB BoolE)
counterSpinWhirlUI = liftM intersectR annulusUI `comp` (counterSpinUI unionR `ap` whirlUI)

-- Not as nifty-looking as I'd hoped.
counterSpinWhirlCropUI :: UI FilterBC
counterSpinWhirlCropUI = liftM cropF counterSpinWhirlUI
 where
   cropF regB im t = crop (regB t) im




annulusUI :: UI Region
annulusUI = liftM2 g (slider "outer radius" (0, 200) 50)
                     (slider "inner radius" (0, 200) 0)
 where
   g outer inner = uscale outer (annulus (inner/outer))


-- Subtraction-based bulging

bulgeSubing :: FloatE -> TransformE
bulgeSubing b = polarXf $ \ (r,a) -> (r-b,a)

bulgeSub :: FloatE -> Filter a
bulgeSub b im = im . bulgeSubing b

bulgeSubUI :: UI (Filter a)
bulgeSubUI = liftM bulgeSub $ slider "bulge" (-50,300) 0

cropBulgeUI :: UI FilterC
cropBulgeUI = cropUI `comp` bulgeSubUI


----------------------------------------

makeAll = makeAllFor allBackends

makeAllFor bends = do
  compileFilterBFor bends "washer filter" $ washerUI
  compileImageBFor  bends "washer bilerp" $ washerBilerpUI
  compileImageBFor  bends "washer tiled"  $ washerTiledUI
  compileImageFor   bends "waves"         $ wavesUI
  compileImageFor   bends "polar checker" $ polarCheckerUI
  compileImageFor   bends "polar checker crop swirl" $ polarCheckerCropSwirlUI
  compileImageBFor  bends "wigglers"      $ wigglersUI
  compileImageFor   bends "swirl fish"    $ swirlFishUI
  compileImageFor   bends "whirl"         $ liftM (flip crop redI) whirlUI
  compileImageBFor  bends "counter-spin whirl" $ counterSpinWhirlUI
  compileFilterFor  bends "bulge subtract" $ atMouse bulgeSubUI
  compileFilterFor  bends "bulge crop"    $ atMouse cropBulgeUI
  compileImageFor   bends "bulge checker" $ liftM ($ cb1) cropBulgeUI
