// hlinkControl.cpp : Defines the entry point for the DLL application.
//

#include "stdafx.h"
#include <windowsx.h>
#include <tchar.h>
#include <malloc.h>
#define EXPORT_DLL
#include "hlinkControl.h"

#define DEFAULT_COLOR  RGB(0,0,255)
#define DEFAULT_VISITCOLOR RGB(128,128,255)
#define DEFAULT_URL TEXT("http://www.microsoft.com/")

#define START_SUB_MESSAGE_MAP(subWProc) \
  WNDPROC wProc = subWProc; \
  \
  switch (uMsg) {
#define HANDLE_MESSAGE(msg, fn) \
   case msg: \
      { BOOL bHandled = TRUE; \
        LRESULT lr = (fn)(hWnd,uMsg,wParam,lParam, &bHandled); \
        if (bHandled) { \
            return lr; \
        } else { \
           return CallWindowProc(wProc,hWnd,uMsg,wParam,lParam); \
  } \
  }
 
#define END_SUB_MESSAGE_MAP() \
    default: \
    return CallWindowProc(wProc,hWnd,uMsg,wParam,lParam); \
    }



BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
    return TRUE;
}

static WNDPROC staticWndProc = NULL;

//
// FUNCTION: drawText
//
// PURPOSE:  Draw the control's text using the specified
//           foreground colour.
//
static
void
drawText(HWND hWnd, COLORREF cText)
{
    LPTSTR staticString;
    DWORD  stringLen;
    PAINTSTRUCT ps;
    HFONT hFont, hFontOld;
    COLORREF cCurrent;
    HDC hDC;
    RECT       windowRect;
    SIZE       txtSize;
    POINT      txtPos = { 0, 0};

    // Sigh - static controls (and a couple of the other std. controls)
    // call upon the parent window to determine what fore- and background
    // colours to use when painting. This leaves the sub-classer kind
    // of stranded, as there's no way to intercept these upcalls asking
    // for what colours to use. (I know question marks could be raised at
    // why you would want to override a system-wide color setting for a
    // control, but there are times when this is The Right Thing to do.)
    // 
    // So, what to do? Well, completely override the message handler that's
    // resp. for these particular upcalls. Three cheers for code reuse.
    //
    // Since we're not able to reuse WM_PAINT code here, there's even less
    // of an incentive to reuse STATIC. But for the moment we will.

    // figure out what text we're supposed to display.

    stringLen = CallWindowProc(staticWndProc, hWnd,WM_GETTEXTLENGTH,0,0);

    staticString = (LPTSTR)_alloca((stringLen + 1)* sizeof(TCHAR));

    CallWindowProc(staticWndProc,hWnd, WM_GETTEXT, (WPARAM)(stringLen+1), (LPARAM)staticString);

    // Let the static control keep track of the HFONT to use.
    hFont = (HFONT)SendMessage(hWnd, WM_GETFONT,0,0);

    hDC = BeginPaint(hWnd,&ps);
    cCurrent = GetTextColor(hDC);

    // only draw the text, not the background.
    SetBkMode(hDC, TRANSPARENT); 

    hFontOld = (HFONT)SelectObject(hDC,hFont);

    SetTextColor(hDC, cText);

    GetClientRect(hWnd,&windowRect);
    GetTextExtentPoint32(hDC,staticString, stringLen, &txtSize);

    DWORD dwExStyle = GetWindowExStyle(hWnd);

    if ((dwExStyle & SS_CENTER) != 0) {
       txtPos.x = windowRect.right / 2 - txtSize.cx / 2;
       txtPos.y = (windowRect.bottom - windowRect.top) / 2 - txtSize.cy / 2;
    } else if ((dwExStyle & SS_RIGHT) != 0) {
       txtPos.x = windowRect.right - txtSize.cx;
       txtPos.y = (windowRect.bottom - windowRect.top) / 2 - txtSize.cy / 2;
    } else {
       txtPos.x = 0;
       txtPos.y = (windowRect.bottom - windowRect.top) / 2 - txtSize.cy / 2;
    }


    TextOut(hDC, txtPos.x, txtPos.y, staticString, stringLen);

    SetTextColor(hDC, cCurrent);
    SelectObject(hDC, hFontOld);

    EndPaint(hWnd,&ps);
}

static
LRESULT
OnSetTarget(HWND hWnd,UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL* bHandled)
{
   LPTSTR str = (LPTSTR)malloc((_tcslen((LPCTSTR)wParam) + 1) * sizeof(TCHAR));
   LPTSTR strOld;

   _tcscpy(str,(LPCTSTR)wParam);

   strOld = (LPTSTR)GetProp(hWnd, TEXT("HLINK_TARGET"));

   if (strOld != NULL) 
      free(strOld);

   SetProp(hWnd, TEXT("HLINK_TARGET"), (HANDLE)str);

   return 0;
}

static
LRESULT
OnSetColor(HWND hWnd,UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL* bHandled)
{
    SetProp
        ( hWnd 
        , (uMsg == HLINK_SETCOLOR ? TEXT("HLINK_COLOR") : TEXT("HLINK_VISITCOLOR"))
        , (HANDLE)wParam
        );
   return 0;

}


    // intercept the font & add underline to it (or, rather, create new
    // font that's underlined.)
static
LRESULT
OnSetFont(HWND hWnd,UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL* bHandled)
{
   HFONT hFont = (HFONT)wParam;
   HFONT hFont2;
   LOGFONT lf;

   GetObject(hFont,sizeof(LOGFONT),&lf);

   if (lf.lfUnderline == TRUE) {
      // optimization - don't create identical font.
      hFont2 = hFont;
   } else {
      lf.lfUnderline = TRUE;
      hFont2 = CreateFontIndirect(&lf);
   }

   // hmm..it'd be better if we had a way of passing back modified WNDPROC
   // params, for the superclass WNDPROC to use, i.e., not hardwire in the
   // name of the WNDPROC.
   return CallWindowProc(staticWndProc,hWnd,uMsg,(WPARAM)hFont2,lParam);
}

// In order to be able receive mouse events from some containers,
// we need to override the STATIC control's handling of the next two.
static
LRESULT
OnGetDlgCode(HWND hWnd,UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL* bHandled)
{
    return DLGC_UNDEFPUSHBUTTON;
}

static
LRESULT
OnHitTest(HWND hWnd,UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL* bHandled)
{
    return HTCLIENT;
}

    // Upon creation, attach the properties holding the URL & colors to use.
static
LRESULT
OnCreate(HWND hWnd,UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL* bHandled)
{
    LPTSTR szDefaultURL = (LPTSTR)malloc ((_tcslen(DEFAULT_URL) + 1) * sizeof(TCHAR));
    _tcscpy(szDefaultURL, DEFAULT_URL);
    SetProp(hWnd, TEXT("HLINK_COLOR"), (HANDLE)DEFAULT_COLOR);
    SetProp(hWnd, TEXT("HLINK_CURRENTCOLOR"), (HANDLE)DEFAULT_COLOR);
    SetProp(hWnd, TEXT("HLINK_VISITCOLOR"), (HANDLE)DEFAULT_VISITCOLOR);
    SetProp(hWnd, TEXT("HLINK_TARGET"), (HANDLE)szDefaultURL);

    // pass message on to superclass.
    *bHandled = FALSE;    
    return 0;
}

static
LRESULT
OnDestroy(HWND hWnd,UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL* bHandled)
{
    // clean up the properties we've attached.
    RemoveProp(hWnd, TEXT("HLINK_COLOR"));
    RemoveProp(hWnd, TEXT("HLINK_CURRENTCOLOR"));
    RemoveProp(hWnd, TEXT("HLINK_VISITCOLOR"));
    free((void*)GetProp(hWnd, TEXT("HLINK_TARGET")));
    RemoveProp(hWnd, TEXT("HLINK_TARGET"));

    // pass message on to superclass.
    *bHandled = FALSE;
    return 0;

}

static
LRESULT
OnPaint(HWND hWnd,UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL* bHandled)
{
    drawText(hWnd,(COLORREF)GetProp(hWnd, TEXT("HLINK_CURRENTCOLOR")));
    return 0;
}

static
LRESULT
OnButtonDown(HWND hWnd,UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL* bHandled)
{
    // in order to be able to see the up message, capture the mouse input.
    SetCapture(hWnd);
    SetProp(hWnd, TEXT("HLINK_CURRENTCOLOR"), (HANDLE)GetProp(hWnd, TEXT("HLINK_VISITCOLOR")));
    InvalidateRect(hWnd, NULL, FALSE);
    return 0;
}

static
LRESULT
OnButtonUp(HWND hWnd,UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL* bHandled)
{
   short x,y;
   RECT r;

   x = GET_X_LPARAM(lParam);
   y = GET_Y_LPARAM(lParam);

   GetClientRect(hWnd, &r);

   if ( (x >= r.left && x <= r.right) &&
        (y >= r.top  && y <= r.bottom) ){
      // follow std behaviour & only consider the interaction
      // for complete when the up event occurs inside the boundaries
      // of the control.
      ShellExecute( NULL
                  , TEXT("open")
                  , (LPCTSTR)GetProp(hWnd, TEXT("HLINK_TARGET"))
                  , NULL
                  , NULL
                  , SW_SHOWNORMAL
                  );
   }
   ReleaseCapture();
            
   // reset color back to non-active (or should we leave it 'visited' ?)
   SetProp(hWnd, TEXT("HLINK_CURRENTCOLOR"), (HANDLE)GetProp(hWnd, TEXT("HLINK_COLOR")));
   InvalidateRect(hWnd, NULL, FALSE);
   return 0;
}

//
// FUNCTION: hlinkWndProc
//
// PURPOSE:  window procedure for hyperlinked label control.
//           All it does is intercept WM_PAINT & adjusts the
//           font to be underlined. The rest of the messages
//           are just forwarded to the STATIC control's WNDPROC.
//
LRESULT
CALLBACK
hlinkWndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    START_SUB_MESSAGE_MAP(staticWndProc)
      HANDLE_MESSAGE(HLINK_SETTARGET, OnSetTarget);
      HANDLE_MESSAGE(HLINK_SETCOLOR, OnSetColor);
      HANDLE_MESSAGE(HLINK_SETVISITCOLOR, OnSetColor);
      HANDLE_MESSAGE(WM_SETFONT, OnSetFont);
      HANDLE_MESSAGE(WM_GETDLGCODE, OnGetDlgCode);
      HANDLE_MESSAGE(WM_NCHITTEST, OnHitTest);
      HANDLE_MESSAGE(WM_NCCREATE, OnCreate);
      HANDLE_MESSAGE(WM_DESTROY, OnDestroy);
      HANDLE_MESSAGE(WM_PAINT, OnPaint);
      HANDLE_MESSAGE(WM_LBUTTONDOWN, OnButtonDown);
      HANDLE_MESSAGE(WM_LBUTTONUP, OnButtonUp);
    END_SUB_MESSAGE_MAP()
}


//
// FUNCTION: ReportLastError(LPCTSTR)
//
// PURPOSE: Simple (and not very user-friendly) way
//          of reporting GetLastError() errors back to
//          the user. For debugging use, only.
//
// COMMENTS: Does not provide support for inserts.
static
void
ReportLastError(LPCTSTR header)
{
    DWORD err = GetLastError();
    LPTSTR msgBuf;

    FormatMessage
            ( FORMAT_MESSAGE_FROM_STRING |
              FORMAT_MESSAGE_ALLOCATE_BUFFER
            , NULL
            , err
            , MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT) // default language
            , (LPTSTR)&msgBuf
            , 0
            , NULL
            );

    MessageBox(NULL, msgBuf, header, MB_OK | MB_ICONINFORMATION);
    LocalFree(msgBuf);
}

//
// FUNCTION: RegisterHLinkClass()
//
// PURPOSE:  Subclass the STATIC control to create HLINK, the
//           window class for a hyperlinked static control.
//           
static
RegisterHLinkClass()
{
    WNDCLASS wc;
    static   initialized = FALSE;

    if (!initialized && GetClassInfo(NULL,TEXT("STATIC"), &wc)) {

      wc.lpszClassName = TEXT("HLINK");
      staticWndProc = wc.lpfnWndProc;
      wc.lpfnWndProc = hlinkWndProc;

      ATOM at = RegisterClass(&wc);
      if (at == (ATOM)0) {
          ReportLastError(TEXT("Failure registering HLink class"));
      }
      
      initialized = TRUE;
    }
}


extern "C"
void
DLLEXPORT
WINAPI
InitHLink()
{
    RegisterHLinkClass();

}
