-- Simple types and literals

module Types where

infixr 9 %->

type Id = String

data Type = Bool | Int | Double | F32vec4 | Fun Type Type | Tuple [Type]
          | Action Type | Ref Type | Arr Type | CType String | Void
  deriving (Eq,Show,Ord)

-- Note: CType constructor is only there to allow us to give precise
--       C types to some parameters of the generated display function.

(%->) :: Type -> Type -> Type
(%->) = Fun

isArrayTy :: Type -> Bool
isArrayTy Arr{} = True
isArrayTy _     = False

isActionTy :: Type -> Bool
isActionTy Action{} = True
isActionTy _	    = False

getActionType :: Type -> Type
getActionType (Action x) = x
getActionType ty = error ("Types.getActionType: not an action type: " ++ show ty)

getRefType :: Type -> Type
getRefType (Ref x) = x
getRefType _ = error "Types.getRefType: not a ref type"

getArrayType :: Type -> Type
getArrayType (Arr x) = x
getArrayType _ = error "Types.getArrayType: not an array"

showCType :: Type -> String
showCType t = 
  case t of
    Bool     -> "int"
    Int      -> "int"
    Double   -> "float"
    F32vec4  -> "F32vec4"
    Action t -> showCType t
    Void     -> "void"
    Arr t    -> showCType t ++ "*"
    Ref t    -> showCType t  -- C doesn't distinguish
    CType s  -> s

-- in bytes 
sizeofType :: Type -> Int
sizeofType t = 
  case t of
    Bool     -> 4
    Int      -> 4
    Double   -> 4  -- nb: rep. as single-precision floats.
    F32vec4  -> 16 -- simd floating point.  4 single-prec floats.
    Action _ -> 4
    Arr t    -> 4  -- ToDo: record size of the array.
    Void     -> 0
    CType s  -> error ("Can't take the size of C type: " ++ show s)
